//#[macro_use] extern crate serde;

use smallvec::SmallVec;
use clap::{
    Parser, Subcommand, ValueEnum,
};
use chain::{
    Chain,
};
use std::{
    num::NonZeroUsize,
    path::{
	PathBuf, Path,
    },
    io::{
	BufRead,
	self,
    },
};

mod iou;

#[derive(Debug, clap::Args)]
#[group(required=false,multiple=true)]
struct ChainFile {
    /// Load the chain from the provided file & then save it back after appending to the chain.
    ///
    /// Note that if there is no data fed to the chain before saving, it is not re-written to the file unless `--force` is also provided.
    ///
    /// # Overwriting
    ///
    /// If `-s` & `-l` are provided on the same file instead of `-a`, but `-f` is not, the save will still fail.
    #[arg(short, long)]
    append: Option<PathBuf>,
    /// Save the chain to the provided output file name after appending to it.
    #[arg(short, long, conflicts_with="append")]
    save: Option<PathBuf>,
    /// Load the chain from the provided output file name before appending to it.
    #[arg(short, long, conflicts_with="append")]
    load: Option<PathBuf>, // TODO: Group with `save` & add `-F/--file` opt for save & load from same file.
}

#[derive(Debug, Parser)]
#[command(name = env!("CARGO_PKG_NAME"), version, about = env!("CARGO_PKG_DESCRIPTION"), long_about = None)]
pub struct Cli {
    #[command(flatten)]
    chain: ChainFile,
    
    /// Force over-writes of files, and force output of binary data to TTY.
    #[arg(short, long)]
    force: bool,

    /// Do not block to read into chain from `stdin` if there is a loaded chain that is not empty.
    ///
    /// If `stdin` is detected as not empty, then it will attempt to consume the buffer still; this prevents the requirement to explicitly close `stdin` when running from a terminal that does not pipe anything into the process input.
    /// To ignore input entirely, use `--write`.
    #[arg(short, long)]
    no_consume: bool,

    /// Do not read into chain from `stdin` ever.
    #[arg(short, long="write")]
    write_only: bool, // XXX: Should we group this with `no_consume`?

    /// The number of lines to output from the chain. Default is 1.
    ///
    /// If 0, generation from chain is skipped.
    #[arg(default_value_t = 1)]
    lines: usize,
}

impl Cli {
    #[inline(always)]
    fn append(&self) -> Option<&Path>
    {
	self.chain.append.as_ref().map(Path::new)
    }
    
    #[inline]
    pub fn load(&self) -> Option<&Path>
    {
	self.chain.load.as_ref().map(Path::new).or_else(|| self.append())
    }
    #[inline] 
    pub fn save(&self) -> Option<&Path>
    {
	self.chain.save.as_ref().map(Path::new).or_else(|| self.append())
    }

    /// If the save operation will be saving to the file loaded by the load operation.
    ///
    /// # Overwriting
    /// Note that forceful overwriting where `-l` & `-s` supply the same file is not applied without `--force`, whereas if `--append` is provided it is.
    #[inline]
    pub fn is_append(&self) -> bool
    {
	self.chain.append.is_some() || (self.chain.save.is_some() && &self.chain.save == &self.chain.load)
    }

    /// If the output should be overwritten.
    pub fn force_overwrite(&self) -> bool
    {
	self.force || self.chain.append.is_some()
    }
}

fn buffered_read_all_lines<T: BufRead+?Sized, F: FnMut(&str) -> io::Result<()>>(input: &mut T, mut then: F) -> io::Result<usize>
{
    let mut buffer = String::new();
    let mut read;
    let mut total=0;
    while {read = input.read_line(&mut buffer)?; read!=0} {
	if buffer.trim().len() > 0 {
	    then(&buffer[..])?;
	}
	buffer.clear();
	total += read;
    }
    Ok(total)
}

#[inline] 
fn parse_cli() -> Cli {
    use clap::Parser;
    Cli::parse()
}

mod format;

fn load_chain<S>(stream: &mut S) -> io::Result<Chain<String>>
where S: io::Read + ?Sized
{
    format::load_chain_from_sync(stream)
}

fn save_chain<S>(stream: &mut S, chain: &Chain<String>) -> io::Result<()>
where S: io::Write + ?Sized
{
    format::save_chain_to_sync(stream, chain, true) //TODO: Change compression to be off for small chains...? We will need to store the chain size info somewhere else.
}

fn create_chain(cli: &Cli) -> Chain<String>
{
    if let Some(load) = cli.load() {
	let mut input = std::fs::OpenOptions::new()
	    .read(true)
	    .open(load).expect("Failed to open chain load file");
	if iou::is_tty(&input) {
	    if cli.force {
		eprintln!("Warning: Reading binary data from TTY!");
	    } else {
		panic!("Cannot read chain from file {:?}: Refusing to read binary data from TTY, if this is intended pass `--force`", load);
	    }
	}
	load_chain(&mut input).expect("Failed to load chain from file")
    } else {
	Chain::new()
    }
}

fn complete_chain(cli: &Cli, chain: Chain<String>, is_unmodified: bool) -> io::Result<()>
{
    if cli.force || // If `--force` is provided, always attempt write.
	!is_unmodified || // And attempt write if there was something appended to the chain
	!cli.is_append() // Or, attempt write if we are *not* appending *even if* there is no change.
    {
	let force = cli.force_overwrite();
	if let Some(save) = cli.save() {
	    let mut output = std::fs::OpenOptions::new()
		.create_new(! force)
		
		.create(force)
		.write(true)
		.truncate(force)
		.open(save).expect("Failed to open chain save file");
	    if iou::is_tty(&output) {
		if cli.force {
		    eprintln!("Warning: Writing binary data to TTY!");
		} else {
		    return Err(io::Error::new(io::ErrorKind::InvalidInput, "Refusing to write binary data to TTY, if this is intended pass `--force`"));
		}
	    }
	    save_chain(&mut output, &chain).expect("Failed to save chain to file") // TODO: Error type
	}
    } else if cfg!(debug_assertions) {
	eprintln!("Warning: Not writing to potential save location {:?}: is_unmodified: {is_unmodified}, is_append: {}, force: overwrite = {} (always: {})", cli.save(), cli.is_append(), cli.force_overwrite(), cli.force);
    }

    Ok(())
}

fn main() {
    let cli = parse_cli();

    let mut unmodified = cli.load().is_some(); //XXX: Should we init this to `false` for all non-loaded chains?
    
    let stdin = io::stdin();
    let mut stdin = stdin.lock();
    let mut chain = create_chain(&cli); // TODO: If `self.force_overwrite() && self.is_append()`, keep the file open for the process duration, and `lseek(0)` it before writing instead of opening it again. (XXX: Should we add `O_EXCL` to the opened file if )

    let has_potential_pending = |stdin| match iou::has_known_available_data(stdin) {
	Err(x) if x.kind() == io::ErrorKind::Unsupported => None,

	Err(err) => {
	    if cfg!(debug_assertions) {
		eprintln!("Warning: Failed to test for pending input: {err}");
	    }
	    None
	},
	Ok(None) => Some(false),
	Ok(Some(_)) => Some(true),
	
    }.unwrap_or(false); // If we cannot see if there is pending data on the stream, assume there is not for the purpose of `no_consume`. (NOTE: `no_consume`'s directive is to only consume *IFF* we **know** there is data to read.)

    if cli.write_only || (cli.no_consume && !(chain.is_empty() || has_potential_pending(&stdin))) {
	if cfg!(debug_assertions) {
	    use std::os::fd::AsFd;
	    eprintln!("Warning: Skipping attempt to read from stdin. write_only: {}, no_consume: {}, is_chain_empty = {}, data_pending_on_p({:?}) = {}", cli.write_only, cli.no_consume, chain.is_empty(), &stdin.as_fd(), has_potential_pending(&stdin));
	}
    } else {
	buffered_read_all_lines(&mut stdin, |string| {
	    let words = string.split_whitespace()
		.filter(|word| !word.is_empty())
		.map(|s| s.to_owned()).collect::<SmallVec<[_; 16]>>();
	    if !words.is_empty() {
		unmodified = false;
		chain.feed(&words);
	    }
	    Ok(())
	}).expect("Failed to read from stdin");
    }

    if !chain.is_empty() {
	match cli.lines {
	    0 => (),
	    1 =>    println!("{}", chain.generate_str()),
	    lines => {
		for string in chain.str_iter_for(lines) {
		    println!("{}", string);
		}
	    },
	};
    }

    complete_chain(&cli, chain, unmodified).expect("Failed to save chain");
}
